import logging
import logging.handlers
from glob import glob
from datetime import datetime, timedelta
import pathlib
import os

SYSLOG_SEVERITIES = [
    'EMERGENCY',
    'ALERT',
    'CRITICAL',
    'ERROR',
    'WARNING',
    'NOTICE',
    'INFORMATIONAL',
    'DEBUG'
]


DEBUG = logging.DEBUG
INFO = logging.INFO
WARNING = logging.WARNING
ERROR = logging.ERROR
CRITICAL = logging.CRITICAL

log_level = {
    'DEBUG': logging.DEBUG,
    "INFO": logging.INFO,
    "WARNiNG": logging.WARNING,
    "ERROR": logging.ERROR,
    "CRITICAL": logging.CRITICAL
}


def create_logger(log_file='', file_level='WARNING', console_level='WARNING', name='', file_mode='a', console=True,
                  syslog=False, syslog_script_name='', log_file_vars=[], log_file_retention_days=0, propagate=False):
    """ Creates a logger and returns the handle.
        Log file vars should be sent as a dict -> {"var": "{date}", "set": "%Y-%m-%d-%Y-%M"}

        Supported log file vars:
            {date} - will be replaced with the current date using the provided strftime format
    
     """
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    logger.propagate = propagate
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    # Console
    if console:
        console_handler = logging.StreamHandler()
        console_handler.setLevel(log_level.get(console_level.upper(), logging.WARNING))
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)

    # syslog
    if syslog:
        syslog_formatter = logging.Formatter(syslog_script_name + '[%(process)d]: %(levelname)s: %(message)s')
        syslog_handler = logging.handlers.SysLogHandler(address='/dev/log')
        syslog_handler.setLevel(log_level.get(file_level.upper(), logging.WARNING))
        syslog_handler.setFormatter(syslog_formatter)
        logger.addHandler(syslog_handler)

    # file
    if log_file != '':
        # replace variables in the log file name
        for var in log_file_vars:
            if type(var) == dict and 'var' in var and 'set' in var and var['var'] == "{date}":
                log_file = log_file.replace(var['var'], datetime.now().strftime(var['set']))
        file_handler = logging.FileHandler(log_file, mode=file_mode, encoding='utf-8', delay=False)
        file_handler.setLevel(log_level.get(file_level.upper(), logging.WARNING))
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

        # manage retention
        if log_file_retention_days > 0:
            # replace variables with a *
            log_file_search_name = log_file
            for var in log_file_vars:
                log_file_search_name.replace(var['var'], '*')
            old_log_files = glob(log_file_search_name)
            for old_log_file in old_log_files:
                # check the age and delete if needed
                fname = pathlib.Path(old_log_file)
                mtime = datetime.fromtimestamp(fname.stat().st_mtime)
                if mtime < datetime.now() - timedelta(days=log_file_retention_days):
                    logger.info('Deleting old log file %s.  Modified time %s, retention set to %i days.', old_log_file, mtime, log_file_retention_days)
                    os.remove(old_log_file)

    # return the logger
    return logger



"""
Class to open and read serial data from the Micropython ammeter
"""
import serial
import logging
from time import sleep, time
from threading import Thread, Lock
from ast import literal_eval


class AmmeterRecvSerial(serial.Serial):
    """ Class to open and read serial data from the Micropython ammeter - overrides the py-serial class """
    def __init__(self, *args, **kwargs):
        self._logger = kwargs.get('logger', logging)
        kwargs.pop('logger')
        super().__init__(*args, **kwargs)
        self.ammeter_data = []
        self.ammeter_partial_data = ''
        self.ammeter_data_lock = Lock()
        self.ammeter_config_lock = Lock()
        self.ammeter_status_lock = Lock()
        self.ammeter_config_data = {}
        self.ammeter_status_data = {}
        self._ammeter_recv_thread = Thread(target=self._ammeter_read, daemon=True)
        self._ammeter_recv_thread.start()

    @property
    def _info_str(self):
        """ Returns a string identifying the class for logging purposes """
        return f"{self.__class__.__name__}: {self.port}: {self.baudrate}: {self.bytesize}{self.parity}{self.stopbits}"

    def ammeter_report(self, wait=5):
        """ Returns the data table collected if the ammeter collection is complete, waits up to the specified time """
        cancel_time = time() + wait
        while time() < cancel_time:
            if self.ammeter_stop_time != {}:
                with self.ammeter_data_lock:
                    return {
                        'start': self.ammeter_start_time,
                        'stop': self.ammeter_stop_time,
                        'data': self.ammeter_data
                    }
        return None

    @property
    def ammeter_status(self):
        """ Get the current status of the ammeter """
        with self.ammeter_status_lock:
            self.ammeter_status_data == {}
        self.write('CMD:STATUS\n'.encode())
        endtime = time() + 3
        while time() < endtime:
            with self.ammeter_status_lock:
                if self.ammeter_status_data != {}:
                    return self.ammeter_status_data
        return {}

    @property
    def ammeter_initialized(self):
        """ Return True/False if ammeter is initialized """
        status = self.ammeter_status
        if status != {} and status['status'] == 'NOINIT':
            return False
        return True

    @property
    def ammeter_running(self):
        """ Return True/ False if sampling is in progress """
        status = self.ammeter_status
        if status != {} and status['status'] == 'RUNNING':
            return True
        return False

    @property
    def ammeter_ready(self):
        status = self.ammeter_status
        if status != {} and status['status'] == 'READY':
            return True
        return False

    @property
    def ammeter_config(self):
        """ Return the current configuration of the ammeter 
            Expecting: CONFIG:{interval}:{timeout}:{pin}:{name}:{baseline}[:{pin}:{name}:{baseline}...]
        """
        with self.ammeter_config_lock:
            self.ammeter_config_data = {}
        self.write('CMD:CONFIG\n'.encode())
        endtime = time() + 3
        while time() < endtime:
            with self.ammeter_config_lock:
                if self.ammeter_config_data != {}:
                    return self.ammeter_config_data
            sleep(.25)
        return {}

    @property
    def ammeter_interval(self):
        """ Return the current sampling interval """
        config = self.ammeter_config
        return config['interval']

    @ammeter_interval.setter
    def ammeter_interval(self, value:int):
        """ Update the sampling interval """
        self.write(f'CMD:INTERVAL:{int(value)}\n'.encode())
        sleep(.2)
        if self.ammeter_interval == value:
            return True
        return False

    def ammeter_init(self):
        """ Initialize the ammeter """
        self.write('CMD:INIT\n'.encode())

    def ammeter_start(self, timeout=None):
        """ Start the sampling """
        self.write((f"CMD:START{(':' + str(timeout)) if timeout is not None else ''}" + '\n').encode())
        stop_time = time() + 3
        while time() < stop_time:
            if self.ammeter_running:
                return True
            sleep(.25)
        return False

    def ammeter_stop(self):
        """ Stop the sampling if currently running """
        self.write('CMD:STOP\n'.encode())
        stop_time = time() + 3
        while time() < stop_time:
            if not self.ammeter_running:
                return True
            sleep(.25)
        return False

    @property
    def ammeter_current(self):
        """ Read one value and return """
        if not self.ammeter_running:
            self.write('CMD:ONE\n'.encode())
            self._ammeter_read()


    def _ammeter_read(self):
        """ Read any data waiting in the queue, if wait_for_response, wait until a non-data response is received """
        self._logger.info('%s: Starting backgroup ammeter read.', self._info_str)
        while True:
            with self.ammeter_data_lock:
                self.ammeter_partial_data += self.read(self.in_waiting).decode('utf-8')
            ammeter_temp_data = self.ammeter_partial_data.split('\n')
            while len(ammeter_temp_data) != 0:
                if self._ammeter_parse_read_line(ammeter_temp_data[0].split(':')):
                    with self.ammeter_data_lock:
                        ammeter_temp_data.pop(0)
                        self.ammeter_partial_data = '\n'.join(ammeter_temp_data)
                else:
                    # didn't find a complete line, if there are other lines then it must be corrupt
                    if len(ammeter_temp_data) > 1:
                        self._logger.warning('%s: Bad data: %s', self._info_str, ammeter_temp_data[0])
                        ammeter_temp_data.pop(0)
                        self.ammeter_partial_data = '\n'.join(ammeter_temp_data)
                    else:
                        # otherwise break out of the loop to wait for more data
                        with self.ammeter_data_lock:
                            self.ammeter_partial_data = '\n'.join(ammeter_temp_data)
                        break
            sleep(.1)


    def _ammeter_parse_read_line(self, response:list):
        """ Take a split list of data recieved from the ammeter and match it to a record """
        if response == ['']:
            # just remove any blank lines
            return True
        self._logger.debug('%s: Received data: %s', self._info_str, response)
        # if received a data point, log it to the array
        if response[0] == 'DATA':
            if len(response) == 6:
                with self.ammeter_data_lock:
                    self.ammeter_data.append({
                        'received': time(),
                        'name': response[1],
                        'ticks': int(response[2]),
                        'current_amps': float(response[3]),
                        'last_reads': literal_eval(response[4]),
                        'average': float(response[5])
                    })
                return True
            else:
                self._logger.error('%s: Invalid number of fields in DATA Transmission.  Received %i, expected 6', self._info_str, len(response))
        # if received a start response, udpate the start or end time
        if response[0] == 'START':
            if len(response) == 2:
                with self.ammeter_data_lock:
                    # clear the data table and set the start time
                    self.ammeter_data = []
                    starttime = time()
                    self.ammeter_start_time = {
                        'reported': int(response[1]),
                        'local': starttime
                    }
                    self.ammeter_stop_time = {}
                return True
            else:
                self._logger.error('%s: Invalid number of fields in START Transmission.  Received %i, expected 2', self._info_str, len(response))
        # if received a stop response, udpate the start or end time
        if response[0] == 'STOP':
            if len(response) == 2:
                with self.ammeter_data_lock:
                    # clear the data table and set the start time
                    stoptime = time()
                    self.ammeter_stop_time = {
                        'reported': int(response[1]),
                        'local': stoptime,
                        'runtime_reported': int(response[1]) - self.ammeter_start_time['reported'],
                        'runtime_local': stoptime - self.ammeter_start_time['local']
                    }
                return True
            else:
                self._logger.error('%s: Invalid number of fields in START Transmission.  Received %i, expected 2', self._info_str, len(response))
        # if received a config repsonse, update the current config
        if response[0] == 'CONFIG':
            if len(response) >= 7:
                with self.ammeter_config_lock:
                    self.ammeter_config_data = {
                        'interval': int(response[1]),
                        'timeout': int(response[2]),
                        'init_timeout': int(response[3]),
                        'pins': [
                            {
                                'pin': int(response[4]),
                                'name': response[5],
                                'baseline': int(response[6])
                            }
                        ]
                    }
                    if len(response) > 7 and (len(response) - 7) % 3 == 0:
                        for x in range(6, len(response), 3):
                            self.ammeter_config_data['pins'].append({})
                return True
            else:
                self._logger.error('%s: Invalid number of fields in CONFIG Transmission.  Received %i, expected >=6', self._info_str, len(response))
        # if received a status response, update the current status
        if response[0] == 'STATUS':
            if len(response) >= 2:
                with self.ammeter_status_lock:
                    self.ammeter_status_data = {
                        'status': response[1],
                        'timeout': response[2] if response[1] == 'INITIALIZING' or response[1] == 'RUNNING' else 0,
                        'noinit_pin': response[2] if response[1] == 'NOINIT' else None
                    }
                return True
            else:
                self._logger.error('%s: Invalid number of fields in STATUS Transmission.  Received %i, expected >=2', self._info_str, len(response))

        # didn't find a match, might not be a complete line
        return False



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ammeter_logger",
    version="0.1.2",
    author="Thomas Dunteman",
    author_email="ammeter@learningtopi.com",
    description="Ammeter logging tool - used to receive from micropython serial connection",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/learningtopi/ammeter_logger",
    project_urls={
        "Bug Tracker": "https://github.com/learningtopi/ammeter_logger/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    install_requires=[
        'pyserial>=3.5'
    ]
)



from .ammeter_recv import AmmeterRecvSerial
from .logging_handler import create_logger
import argparse
from time import sleep, time
import json
import csv
from datetime import datetime
import signal

def write_log_data(serial_device:AmmeterRecvSerial, args:dict):
    """ Write the log data to the specified file """
    print(f"Writing log to file {args['file']}...")
    with open(args['file'], 'w', encoding='utf-8') as output_file:
        writer = csv.writer(output_file)
        header = ['received_epoch', 'name', 'received_datetime', 'ticks', 'latest']
        for x in range(len(serial_device.ammeter_data[0]['last_reads'])):
            header.append(f'read_{x+1}')
        header.append('time_from_start')
        header.append('average')
        writer.writerow(header)
        for record in serial_device.ammeter_data:
            row = [
                record['received'], 
                record['name'],  
                datetime.fromtimestamp(record['received']).strftime('%Y-%m-%d %H:%M:%S.%f'),
                record['ticks'], 
                record['current_amps']]
            for x in range(len(record['last_reads'])):
                row.append(record['last_reads'][x])
            row.append(float(record['ticks'])/1000)
            row.append(record['average'])
            writer.writerow(row)
    print("Writing complete.")


if __name__ == '__main__':
    # setup the argument parser
    parser = argparse.ArgumentParser(description="Start the ammeter data collector.  Requires the sender to be running (provided sender is Micropython for a microcontroller)")
    parser.add_argument('device', metavar='DEVICE', help='Serial device connected to the microcontroller (i.e. /dev/ttyUSB0')
    parser.add_argument('file', metavar='OUTPUT_FILE', help='File to save captured data to')
    parser.add_argument('--get-config', dest='get_config', required=False, action='store_true', default=False, help="(False) Get the configuration from the microcontroller and quit")
    parser.add_argument('--get-status', dest='get_status', required=False, action='store_true', default=False, help="(False) Get the current status of the microcontroller and quit")
    parser.add_argument('--skip-init', dest='skip_init', required=False, action='store_true', default=False, help="(False) Skip initializing the ammeter (not recommended!)")
    parser.add_argument('--force-init', dest='force_init', required=False, action='store_true', default=False, help="(False) Force init of the ammeter")
    parser.add_argument('--init-only', dest='init_only', required=False, action='store_true', default=False, help="(False) Only initalize the ammeter, print the config and status and quit (implies --force-init)")
    parser.add_argument('--sample-interval', dest='sample_interval', required=False, type=int, default=0, help="Set the sampling interval, overrides the config on the microcontroller")
    parser.add_argument('--capture-time', dest='capture_time', required=False, type=int, default=None, help="Set the max time to capture before stopping, overrides the config on the microcontroller")
    parser.add_argument('--baudrate', dest='baudrate', required=False, type=int, default=115200, help="(115200) Set the baudrate for the serial interface")
    parser.add_argument('--log-level', dest='log_level', required=False, type=str, default='INFO', help='(INFO) Specify the logging level for the console')

    args = vars(parser.parse_args())

    # Create the device
    serial_device = AmmeterRecvSerial(args['device'], baudrate=args['baudrate'], logger=create_logger(console=True, console_level=args['log_level']))

    # get config and status if requested and quit
    if args['get_config']:
        print(f"Current Config: {serial_device.ammeter_config}")
        if not args['get_status']:
            quit()
    if args['get_status']:
        print(f"Current Status: {serial_device.ammeter_status}")
        quit()

    # initialize the device
    if (not serial_device.ammeter_initialized and not args['skip_init']) or args['force_init'] or args['init_only']:
        serial_device.ammeter_init()
        timeout = time() + 2 + serial_device.ammeter_config['init_timeout']
        while time() < timeout:
            status = serial_device.ammeter_status
            if status['status'] != 'READY' or status['status'] != 'RUNNING':
                print(f'Waiting for ammeter to initialize. {status}')
            else:
                break
            sleep(2)
        if not serial_device.ammeter_ready:
            print(f"Ammeter is not ready!  Current status: {serial_device.ammeter_status}")
            quit(1)
    if args['init_only']:
        print(f"Current Config: {serial_device.ammeter_config}")
        print(f"Current Status: {serial_device.ammeter_status}")
        quit()

    # start the logging
    return_value = serial_device.ammeter_start(timeout=args['capture_time'])
    if not return_value:
        print(f"Error starting the ammeter!  Current status: {serial_device.ammeter_status}")
        quit()
    
    # Configure a handler to catch a Ctrl+C
    def break_handler(signum, frame):
        """ Handle a ctrl+c from the user to stop the data collection and write the collected data """
        print(f"Caught Ctrl+C.  Stopping data collection...")
        if serial_device.ammeter_stop():
            write_log_data(serial_device, args)
        else:
            print(f"Error stopping the ammeter!  Current status: {serial_device.ammeter_status}")
        quit()
    signal.signal(signal.SIGINT, break_handler)

    # wait for the logging to complete
    print(f"Starting data collection.  Collection will run for {serial_device.ammeter_config['timeout']} seconds.  You can stop at any point and write the captured data using CTRL+C.")
    timeout = time() + 2 + (serial_device.ammeter_config['timeout'] if args['capture_time'] is None else args['capture_time'])
    while time() < timeout:
        status = serial_device.ammeter_status
        if status['status'] == 'RUNNING':
            last_read = serial_device.ammeter_data[len(serial_device.ammeter_data) - 1]['average'] if len(serial_device.ammeter_data) > 0 else 0
            print(f"Waiting for logging run to complete.  Last amp read: {last_read}.  Current status: {status}")
        elif status['status'] == 'READY':
            print(f"Logging Run complete.  Current status: {status}.  Captured {len(serial_device.ammeter_data)} intervals")
            break
        else:
            print(f"Microcontroller reporting an unexpected state: {status}.  Check and try again.")
            quit(1)
        sleep(2)
    
    # logging complete, write the data to a file
    write_log_data(serial_device, args)